import concurrent.futures
import hashlib
import pathlib
from typing import Tuple

import numpy as np

from config import MAX_BIOS_BYTES


def hash_file(file_path: pathlib.Path) -> str:
    """

    :param file_path:
    :return:
    """

    file_bytes = file_path.read_bytes()
    file_hash = hashlib.md5(file_bytes)
    return file_hash.hexdigest()


def recursive_hash(
    directory: pathlib.Path, glob: str = "*"
) -> Tuple[np.ndarray, np.ndarray]:
    """

    :param directory:
    :param glob:
    :return:
    """

    file_paths = list(directory.rglob(pattern=glob))
    file_paths = [
        file_path
        for file_path in file_paths
        if file_path.stat().st_size <= MAX_BIOS_BYTES and file_path.is_file()
    ]

    file_hashes = []
    with concurrent.futures.ThreadPoolExecutor() as executor:
        for file_hash in executor.map(hash_file, file_paths):
            file_hashes.append(file_hash)
    return np.array(file_paths), np.array(file_hashes)


def match_arrays(
    array_a: np.ndarray, array_b: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Element-wise array comparison, returns unique values and matching indices per input array

    :param array_a:
    :param array_b:
    :return:
    """

    # expecting 1D arrays
    if np.sum([len(array_a.shape), len(array_b.shape)]) > 2:
        raise ValueError("input arrays need to be one-dimensional, exiting..")

    comparisons = np.equal(array_a.reshape(1, -1), array_b.reshape(-1, 1))

    indices_b, indices_a = np.where(comparisons)
    matching_values = np.unique(array_a[indices_a])

    return matching_values, indices_a, indices_b



import pathlib
import re
import urllib.request

import pandas as pd

# not expecting BIOS files over 15mb
MAX_BIOS_BYTES = 15728640
SEED = 0

# Pulling all BIOS names and hashes from Libretro's system.dat (https://docs.libretro.com/)
FILE_PATH = pathlib.Path(__file__).parent / "system.dat"
GITHUB_URL = (
    "https://raw.githubusercontent.com/libretro/libretro-database/master/dat/System.dat"
)
if not FILE_PATH.exists():
    print("Getting BIOS names from libretro-database..")
    urllib.request.urlretrieve(GITHUB_URL, FILE_PATH)
    print("Done.")

# Parsing Libretro's system.dat and formatting as pandas dataframe
index = 0 # pylint: disable=invalid-name
SYSTEMS = []
with open(FILE_PATH, "r", encoding='utf-8') as file:
    for line in file:
        line = line.strip()
        if line.startswith("comment"):
            current_system = line.split('"')[1]
        elif line.startswith("rom"):
            match = re.search(
                r"name (\S+)(?: size (\S+))?(?: crc (\S+))?(?: md5 (\S+))?(?: sha1 (\S+))?",
                line,
            )
            data = {
                "system": current_system,
                "name": match.group(1).replace('"', "").replace("'", ""),
                "size": match.group(2) if match.group(2) else None,
                "crc": match.group(3) if match.group(3) else None,
                "md5": match.group(4) if match.group(4) else None,
                "sha1": match.group(5) if match.group(5) else None,
            }
            SYSTEMS.append(pd.DataFrame(data, index=[index]))
            index += 1

# join dfs and drop features without checksums
SYSTEMS = pd.concat(SYSTEMS)
SYSTEMS = SYSTEMS[~SYSTEMS["md5"].isnull()].reset_index(drop=True)



import argparse
import shutil
import pathlib
import numpy as np
import tqdm

from config import SYSTEMS as system_df
from libretro_finder.utils import match_arrays, recursive_hash


def organize(
    search_dir: pathlib.Path, output_dir: pathlib.Path, glob: str = "*"
) -> None:
    """
    Non-destructive function that finds, copies and refactors files to the format expected by
    libretro (and its cores). This is useful if you source your BIOS files from many different
    places and have them saved them under different names (often with duplicates).

    :param search_dir:
    :param output_dir:
    :param glob:
    :param overwrite:
    :return:
    """

    # Indexing files to be checked for matching MD5 checksums
    output_dir.mkdir(parents=True, exist_ok=True)
    file_paths, file_hashes = recursive_hash(directory=search_dir, glob=glob)

    # Element-wise matching of files against libretro's files
    matching_values, file_indices, system_indices = match_arrays(
        array_a=file_hashes, array_b=np.array(system_df["md5"].values)
    )

    if not np.size(matching_values) > 0:
        print("No matching BIOS files were found, exiting..")
        return

    # If a single core has multiple matches, we just pick the first one
    _, indices = np.unique(system_indices, return_index=True)
    srcs = file_paths[file_indices[indices]]
    hashes = file_hashes[file_indices[indices]]
    system_subset = system_df.loc[system_indices[indices]]

    # np.unique and indexing doesn't merit a dedicated function but it should still be tested
    assert np.array_equal(np.array(system_subset["md5"].values), hashes)
    assert system_subset["name"].size == system_subset["name"].unique().size

    # printing matches per system
    matches = system_subset.groupby("system").count()
    print(
        f"{matches['name'].sum()} matching BIOS files were found for {matches.shape[0]}"
        "unique systems:"
    )
    for name, row in matches.iterrows():
        print(f"\t{name} ({row['name']})")

    # copying matching files to output_dir using structure specified by libretro
    dsts = system_subset["name"].values

    # checking whether our input and output paths are of equal length
    assert len(srcs) == len(dsts)

    for i in tqdm.tqdm(range(srcs.size), total=srcs.size):
        dst = output_dir / dsts[i]
        parent = dst.parent
        if dst.exists() or srcs[i] == dst:
            continue
        if parent != output_dir:
            parent.mkdir(parents=True, exist_ok=True)

        shutil.copy(src=srcs[i], dst=dst)


def main() -> None:
    """Simple argparse wrapper for packaging."""
    parser = argparse.ArgumentParser(
        description="CLI that finds, copies and refactors BIOS files "
        "to the format expected by libretro (i.e. name and directory structure).",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("search_dir", help="Directory to look for BIOS files", type=str)
    parser.add_argument(
        "output_dir", help="Directory to refactor found BIOS files to", type=str
    )
    parser.add_argument(
        "-g",
        "--glob",
        help="Glob pattern to use for file matching",
        type=str,
        default="*",
    )
    args = vars(parser.parse_args())

    search_directory = pathlib.Path(args["search_dir"])
    output_directory = pathlib.Path(args["output_dir"])
    search_glob = args["glob"]

    organize(search_dir=search_directory, output_dir=output_directory, glob=search_glob)


if __name__ == "__main__":
    main()
